package javax.activation;

import java.util.Enumeration;
import java.util.Locale;
import java.util.Hashtable;

public class MimeTypeParameterList
{
    private Hashtable parameters;
    private static final String TSPECIALS = "()<>@,;:/[]?=\\\"";
    
    public MimeTypeParameterList() {
        this.parameters = new Hashtable();
    }
    
    public MimeTypeParameterList(String parameterList) throws MimeTypeParseException {
        this.parameters = new Hashtable();
        this.parse(parameterList);
    }
    
    protected void parse(String parameterList) throws MimeTypeParseException {
        if (parameterList == null) {
            return;
        }
        int length = parameterList.length();
        if (length <= 0) {
            return;
        }
        int i = skipWhiteSpace(parameterList, 0);
        char c;
        while (i < length && (c = parameterList.charAt(i)) == ';') {
            ++i;
            i = skipWhiteSpace(parameterList, i);
            if (i >= length) {
                return;
            }
            int lastIndex = i;
            while (i < length && isTokenChar(parameterList.charAt(i))) {
                ++i;
            }
            String name = parameterList.substring(lastIndex, i).toLowerCase(Locale.ENGLISH);
            i = skipWhiteSpace(parameterList, i);
            if (i >= length || parameterList.charAt(i) != '=') {
                throw new MimeTypeParseException("Couldn't find the '=' that separates a parameter name from its value.");
            }
            ++i;
            i = skipWhiteSpace(parameterList, i);
            if (i >= length) {
                throw new MimeTypeParseException("Couldn't find a value for parameter named " + name);
            }
            c = parameterList.charAt(i);
            String value;
            if (c == '\"') {
                if (++i >= length) {
                    throw new MimeTypeParseException("Encountered unterminated quoted parameter value.");
                }
                lastIndex = i;
                while (i < length) {
                    c = parameterList.charAt(i);
                    if (c == '\"') {
                        break;
                    }
                    if (c == '\\') {
                        ++i;
                    }
                    ++i;
                }
                if (c != '\"') {
                    throw new MimeTypeParseException("Encountered unterminated quoted parameter value.");
                }
                value = unquote(parameterList.substring(lastIndex, i));
                ++i;
            }
            else if (isTokenChar(c)) {
                lastIndex = i;
                while (i < length && isTokenChar(parameterList.charAt(i))) {
                    ++i;
                }
                value = parameterList.substring(lastIndex, i);
            }
            else {
                throw new MimeTypeParseException("Unexpected character encountered at index " + i);
            }
            this.parameters.put(name, value);
            i = skipWhiteSpace(parameterList, i);
        }
        if (i < length) {
            throw new MimeTypeParseException("More characters encountered in input than expected.");
        }
        return;
    }
    
    public int size() {
        return this.parameters.size();
    }
    
    public boolean isEmpty() {
        return this.parameters.isEmpty();
    }
    
    public String get(String name) {
        return (String)this.parameters.get(name.trim().toLowerCase(Locale.ENGLISH));
    }
    
    public void set(String name, String value) {
        this.parameters.put(name.trim().toLowerCase(Locale.ENGLISH), value);
    }
    
    public void remove(String name) {
        this.parameters.remove(name.trim().toLowerCase(Locale.ENGLISH));
    }
    
    public Enumeration getNames() {
        return this.parameters.keys();
    }
    
    @Override
    public String toString() {
        StringBuffer buffer = new StringBuffer();
        buffer.ensureCapacity(this.parameters.size() * 16);
        Enumeration<Object> keys = this.parameters.keys();
        while (keys.hasMoreElements()) {
            String key = (String)keys.nextElement();
            buffer.append("; ");
            buffer.append(key);
            buffer.append('=');
            buffer.append(quote((String)this.parameters.get(key)));
        }
        return buffer.toString();
    }
    
    private static boolean isTokenChar(char c) {
        return c > ' ' && c < '\u007f' && "()<>@,;:/[]?=\\\"".indexOf((int)c) < 0;
    }
    
    private static int skipWhiteSpace(String rawdata, int i) {
        for (int length = rawdata.length(); i < length && Character.isWhitespace(rawdata.charAt(i)); ++i) {}
        return i;
    }
    
    private static String quote(String value) {
        boolean needsQuotes = false;
        int length = value.length();
        for (int i = 0; i < length && !needsQuotes; needsQuotes = !isTokenChar(value.charAt(i)), ++i) {}
        if (needsQuotes) {
            StringBuffer buffer = new StringBuffer();
            buffer.ensureCapacity((int)((double)length * 1.5));
            buffer.append('\"');
            for (int j = 0; j < length; ++j) {
                char c = value.charAt(j);
                if (c == '\\' || c == '\"') {
                    buffer.append('\\');
                }
                buffer.append(c);
            }
            buffer.append('\"');
            return buffer.toString();
        }
        return value;
    }
    
    private static String unquote(String value) {
        int valueLength = value.length();
        StringBuffer buffer = new StringBuffer();
        buffer.ensureCapacity(valueLength);
        boolean escaped = false;
        for (int i = 0; i < valueLength; ++i) {
            char currentChar = value.charAt(i);
            if (!escaped && currentChar != '\\') {
                buffer.append(currentChar);
            }
            else if (escaped) {
                buffer.append(currentChar);
                escaped = false;
            }
            else {
                escaped = true;
            }
        }
        return buffer.toString();
    }
}
